from __future__ import (absolute_import, division,print_function)
import maya.cmds as cmds
import os
from importlib import import_module

try:
    from importlib import reload
except:
    pass

from . import rigUtil,rigSettings
reload(rigUtil)
reload(rigSettings)

##-------------------------------------------------------------------------------
## for rigModule command
##-------------------------------------------------------------------------------
##リグモジュール置き場
RIGMODULELIB = "SSRig.rigModules"

def getModuleInfo(moduleName):
    ##リグモジュールをimport
    rigModule = import_module("."+moduleName ,RIGMODULELIB)
    reload(rigModule)

    ##アトリビュートを問い合わせ
    attrs = list(rigModule._ATTRS)
    attrTypes = dict(rigModule._ATTR_DICT)
    moduleType = str(rigModule._MODULETYPE)
    return moduleType,attrs,attrTypes

def addAttrs(moduleNode,attr,attrDict,opts):
    if cmds.attributeQuery(attr, node = moduleNode, exists =True):
        return

    if attrDict[attr]["type"] == "string":
        cmds.addAttr(moduleNode,ln = attr, dt = "string", multi = attrDict[attr]["multi"],**opts)
        
        if attrDict[attr]["multi"] == False:
            cmds.setAttr(moduleNode + "."+attr,attrDict[attr]["default"], type ="string",**opts)
    
    elif attrDict[attr]["type"] == "double":
        cmds.addAttr(moduleNode,ln = attr, at = "double", dv = attrDict[attr]["default"],min = attrDict[attr]["min"],max = attrDict[attr]["max"],**opts)

    elif attrDict[attr]["type"] == "long":
        cmds.addAttr(moduleNode,ln = attr, at = "long", dv = attrDict[attr]["default"],min = attrDict[attr]["min"],max = attrDict[attr]["max"],**opts)

    elif attrDict[attr]["type"] == "message":
        cmds.addAttr(moduleNode,ln = attr, at = "message", multi = attrDict[attr]["multi"],**opts)

    elif attrDict[attr]["type"] == "enum":
        enumString = ":".join(attrDict[attr]["enumList"])
        cmds.addAttr(moduleNode,ln = attr, at = "enum", enumName = enumString, dv = attrDict[attr]["enumList"].index(attrDict[attr]["default"]),**opts)

    elif attrDict[attr]["type"] == "bool":
        cmds.addAttr(moduleNode,ln = attr, at = "bool", dv = attrDict[attr]["default"],**opts)

    elif attrDict[attr]["type"] == "compound":
        cmds.addAttr(moduleNode, ln = attr, at = "compound", numberOfChildren = len(attr["ch"]), multi = attr["multi"],**opts)
        
        for chAttr in attr["ch"]:
            addAttrs(moduleNode,chAttr,{"p":attr},attrDict)

def getValues(moduleNode,attr,parentAttr,attrDict):
    ##アトリビュートが無い場合は設定したデフォルト値を採用
    if cmds.attributeQuery(attr, node = moduleNode, exists =True) ==False:
        return attrDict[attr]["default"]

    ##アトリビュートのタイプによって処理を変更
    if attrDict[attr]["type"] == "message":
        ##message
        if attrDict[attr]["multi"]:
            return cmds.listConnections(moduleNode + "."+ parentAttr + attr,s=True,d=False) or []
        
        else:
            inputs = cmds.listConnections(moduleNode + "."+ parentAttr + attr,s=True,d=False) or [""]
            return inputs[0]
        
    elif attrDict[attr]["type"] == "string":
        ##string
        if attrDict[attr]["multi"]:
            inputs = []
            for index in  cmds.getAttr(moduleNode +"."+ parentAttr + attr, mi =True) or []:
                inputs.append(cmds.getAttr(moduleNode +"."+ parentAttr + attr + "["+str(index)+"]") or "")            
            return inputs
        else:
            return cmds.getAttr(moduleNode + "."+ parentAttr + attr) or ""

    elif attrDict[attr]["type"] == "enum":
        ##enum
        index = cmds.getAttr(moduleNode +"."+ attr)
        return cmds.attributeQuery(attr, node = moduleNode, listEnum =True)[0].split(":")[index]

    elif attrDict[attr]["type"] == "compound":
        compoundValues = []        
        if attrDict[attr]["multi"]:
            indexes = cmds.getAttr(moduleNode +"."+ attr,mi =True) or []            
            for i in indexes:
                valueDict = {}
                parentAttr = attr + "["+str(i)+"]."
                for chAttr in attrDict[attr]["ch"]:
                    chValue = getValues(moduleNode,chAttr,parentAttr,attrDict)
                    valueDict[chAttr] = chValue
                compoundValues.append(valueDict)
            return compoundValues
        
        else:
            valueDict = {}
            for chAttr in attrDict[attr]["ch"]:
                chValue = getValues(moduleNode,chAttr,parentAttr,attrDict)
                valueDict[chAttr] = chValue                
            return valueDict
    else:
        ##bool / long / double
        return cmds.getAttr(moduleNode + "." + attr)

def setValues(moduleNode,attr,parentAttr,attrDict,value):        
    if cmds.objExists(moduleNode +"."+ parentAttr + attr) == False:
        return

    if cmds.getAttr(moduleNode +"."+ parentAttr + attr, l =True) == True:
        return

    if attrDict[attr]["type"] == "string":
        if attrDict[attr]["multi"]:
            for i in range(0,len(value)):
                cmds.setAttr(moduleNode +"."+ parentAttr + attr + "["+str(i)+"]",value[i],type = "string")
        else:
            cmds.setAttr(moduleNode +"."+ parentAttr + attr, value,type = "string")
    
    elif attrDict[attr]["type"] == "double" or attrDict[attr]["type"] == "long":
        cmds.setAttr(moduleNode +"."+ parentAttr + attr,value)

    elif attrDict[attr]["type"] == "message":
        if attrDict[attr]["multi"]:
            for i in range(0,len(value)):                
                if value[i] != "":    
                    cmds.connectAttr(value[i] + ".message",moduleNode +"."+ parentAttr + attr + "["+str(i)+"]",f =True)
        
                else:
                    curInput = cmds.listConnections(moduleNode +"."+ parentAttr + attr, s =True, d=False,p=True)
                    if curInput != None:
                        cmds.disconnectAttr(curInput[0],moduleNode +"."+ parentAttr + attr)
        else:
            if value != "":
                cmds.connectAttr(value + ".message",moduleNode +"."+ parentAttr +attr,f =True)
        
            else:
                curInput = cmds.listConnections(moduleNode +"."+ parentAttr + attr, s =True, d=False,p=True)
                if curInput != None:
                    cmds.disconnectAttr(curInput[0],moduleNode +"."+ parentAttr + attr)

        
    elif attrDict[attr]["type"] == "enum":        
        cmds.setAttr(moduleNode +"."+ parentAttr + attr,attrDict[attr]["enumList"].index(value))

    elif attrDict[attr]["type"] == "bool":
        cmds.setAttr(moduleNode +"."+ parentAttr + attr, value)

    elif attrDict[attr]["type"] == "compound":        
        if attrDict[attr]["multi"]:
            for i in range(0,len(value)):
                parentAttr = attr + "["+str(i)+"]."
                for chParam in list(value[i].keys()):
                    setValues(chParam,moduleNode,value[i][chParam],parentAttr,attrDict)        
        else:            
            parentAttr = attr+"."
            for chParam in list(value.keys()):
                setValues(chParam,moduleNode,value[chParam],parentAttr,attrDict)

##excute------------
def getModuleNodeValues(moduleNode):
    valueDict = {}
    moduleName = cmds.getAttr(moduleNode + ".moduleName") 
    moduleType,attrs,attrDict = getModuleInfo(moduleName)
    
    ##共通値取得
    commonAttrDict = rigSettings.COMMON_ATTR_DICT[moduleType]
    for attr in list(commonAttrDict.keys()):
        valueDict[attr] = getValues(moduleNode,attr,"",commonAttrDict)

    ##固有値取得
    for attr in attrs:
        valueDict[attr] = getValues(moduleNode,attr,"",attrDict)

    return valueDict

def execModuleCmd(moduleName,valueDict):
    ##リグモジュールをimport
    rigModule = import_module("."+moduleName ,RIGMODULELIB)
    reload(rigModule)    
    
    ##セットアップを実行
    result = rigModule.main(**valueDict)
    return result

def excuteModule(moduleNode,baseOpt):
    ##使用するリグモジュール名を取得
    moduleName = cmds.getAttr(moduleNode + ".moduleName") 
    
    ##ノードからリグモジュールに渡す各値を取得
    valueDict = getModuleNodeValues(moduleNode)

    ##リグモジュールによるセットアップを実行
    valueDict.update(**baseOpt)
    execModuleCmd(moduleName,valueDict)

def excuteSceneModules(topNode,baseOpt):
    if cmds.objExists(topNode):
        moduleNodes = cmds.listRelatives(topNode, type = "transform") or []
        
        ##sortOrder
        moduleSortedDict = {}
        for node in moduleNodes:
            if cmds.attributeQuery("moduleName", node = node, exists =True) == False:
                continue

            orderNum = cmds.getAttr(node + ".orderOfExecution")
            
            if orderNum not in list(moduleSortedDict.keys()):
                moduleSortedDict[orderNum] = [node]
            else:
                moduleSortedDict[orderNum].append(node)

        for key in sorted(list(moduleSortedDict.keys())):
            for moduleNode in sorted(moduleSortedDict[key]):
                excuteModule(moduleNode,baseOpt)

##create--------------------------
def createModuleNode(section,prefix,side,moduleName):
    ##リグモジュールノード名を生成
    moduleNodeName = rigUtil.addPrefix(section,prefix) + rigUtil.capitalizeString(moduleName)+ rigUtil.getSideString(side) + rigSettings.SUFFIX_DICT["module"]

    ##リグモジュールノードグループを作成
    if cmds.objExists(rigSettings.NODENAME_DICT["moduleGP"]) ==False:
        cmds.createNode("transform", name = rigSettings.NODENAME_DICT["moduleGP"])

    moduleNode = cmds.createNode("transform",name = moduleNodeName, p = rigSettings.NODENAME_DICT["moduleGP"])

    ##リグモジュールに問い合わせ
    moduleType,attrs,attrDict = getModuleInfo(moduleName)

    ##必須アトリビュートを先に追加
    if moduleType in list(rigSettings.COMMON_ATTR_DICT.keys()):
        commonAttrDict = rigSettings.COMMON_ATTR_DICT[moduleType]
        
        for attr in list(commonAttrDict.keys()):
            addAttrs(moduleNode,attr,commonAttrDict,{})
    
        ##必須アトリビュート値設定
        setValues(moduleNode,"moduleName","",commonAttrDict,moduleName)
        setValues(moduleNode,"section","",commonAttrDict,section)
        setValues(moduleNode,"prefix","",commonAttrDict,prefix)
        setValues(moduleNode,"side","",commonAttrDict,side)
        
    ##リグモジュール固有アトリビュートを追加
    for attr in attrs:
        addAttrs(moduleNode,attr,attrDict,{})

    return moduleNode

def updateModuleNode(moduleNode):
    #現在の値を取得
    valueDict = getModuleNodeValues(moduleNode)
    
    #旧ノードをリネーム
    cmds.parent(moduleNode,w=True)
    oldNode = cmds.rename(moduleNode, moduleNode + "_old")

    #新ノードを作成
    moduleNode = createModuleNode(valueDict["section"],valueDict["prefix"],valueDict["side"],valueDict["moduleName"])
    moduleType,attrs,attrDict = getModuleInfo(valueDict["moduleName"])

    ##旧ノードの値をセット
    for attr in attrs:
        if attr in list(valueDict.keys()):
            setValues(moduleNode,attr,"",attrDict,valueDict[attr])
        
    chNodes = cmds.listRelatives(oldNode,type = "transform") or None
    if chNodes != None:
        cmds.parent(chNodes,moduleNode,a =True)

    cmds.select(moduleNode,r=True)

##-------------------------------------------------------------------------------
## for preProcess
##-------------------------------------------------------------------------------
def createNewScene(force = True):
    cmds.file(new=True,f=force)

    ##unknownPlugin 除去
    for plugin in cmds.unknownPlugin(query = True, list =True) or []:
        cmds.unknownPlugin(plugin,r =True)

    ##namespace　除去
    ignorList = ['UI', 'shared']
    allNameSpace = cmds.namespaceInfo(listOnlyNamespaces = True,recurse =True)
    
    for namespace in allNameSpace:
        if namespace in ignorList:
            continue

        cmds.namespace(removeNamespace = namespace)

def importFile(filePath):
    if os.path.exists(filePath) ==False:
        return []
    
    ##インポート前の 最上位DAGノード取得 
    beforeTopNodes = set(cmds.ls(assemblies =True))
    
    ##ファイルのインポート
    cmds.file(
            filePath,
            i = True,
            options = 'v=0;',
            renameAll=False,
            loadReferenceDepth="none"
        )

    ##インポート後の 最上位DAGノード取得
    afterTopNodes = set(cmds.ls(assemblies =True))

    ##インポートによって追加された 最上位DAGノード取得
    addedTopNodes = list(afterTopNodes  - beforeTopNodes)

    return addedTopNodes 

def preProcess(modelFilePath,skeletonFilePath,globalScale,setupDataPath):
    ##000
    createNewScene(True)

    ##001
    modelNodes = importFile(modelFilePath)
    skeletonNodes = importFile(skeletonFilePath)

    baseOpt = {
                "globalScale":globalScale,
                "setupDataPath":setupDataPath
                }

    ##050
    ##createBasicStructure
    execModuleCmd("createBasicStructure",baseOpt)
    cmds.parent(modelNodes,rigSettings.NODENAME_DICT["modelGP"],a=True)

##-------------------------------------------------------------------------------
## for setupProcess
##-------------------------------------------------------------------------------

def setupProcess(globalScale,setupDataPath):
    baseOpt = {
                "globalScale":globalScale,
                "setupDataPath":setupDataPath
                }
    
    ##100~ 
    ##200~ setup
    ##300~
    ##400~
    ##500~ deform
    ##600~ 
    ##700~ ctrl
    ##800~ switch    
    excuteSceneModules(rigSettings.NODENAME_DICT["moduleGP"],baseOpt)

##-------------------------------------------------------------------------------
## for postProcess
##-------------------------------------------------------------------------------
def exportPostTransform(setupDataPath):
    filePath = setupDataPath + rigSettings.POSTFILENAME_DICT["transform"]
   
    postTransformDict = {
                            "translateX":   cmds.getAttr(rigSettings.NODENAME_DICT["ctrlWorldSpace"]+".translateX"),
                            "translateY":   cmds.getAttr(rigSettings.NODENAME_DICT["ctrlWorldSpace"]+".translateY"),
                            "translateZ":   cmds.getAttr(rigSettings.NODENAME_DICT["ctrlWorldSpace"]+".translateZ"),
                            "rotateX":      cmds.getAttr(rigSettings.NODENAME_DICT["ctrlWorldSpace"]+".rotateX"),
                            "rotateY":      cmds.getAttr(rigSettings.NODENAME_DICT["ctrlWorldSpace"]+".rotateY"),
                            "rotateZ":      cmds.getAttr(rigSettings.NODENAME_DICT["ctrlWorldSpace"]+".rotateZ"),
                            "scale":        cmds.getAttr(rigSettings.NODENAME_DICT["ctrlWorldSpace"]+".postScale")
        }

    rigUtil.writeJSON(filePath,postTransformDict,indent = 4)    

def exportAllCtrlValues(setupDataPath):
    allCtrls = rigUtil.listSetMembers(rigSettings.NODENAME_DICT["allCtrlSet"],["transform"],True)
    allCtrlValueDict = {}

    for ctrl in allCtrls:        
        attrs = rigUtil.listAttrs(ctrl,True,True,True)
        attrsDict = {}
        
        for attr in attrs:            
            curValue = cmds.getAttr(ctrl + "."+attr)
            attrsDict[attr] = round(curValue,5)

        limitDict = rigUtil.getTransformLimit(ctrl)
        attrsDict.update(**limitDict)
        
        if len(attrsDict) != 0:
            allCtrlValueDict[ctrl] = attrsDict

    filePath = setupDataPath + rigSettings.POSTFILENAME_DICT["ctrlValue"]
    rigUtil.writeJSON(filePath,allCtrlValueDict,indent = 4)

def exportAllCtrlCV(setupDataPath):
    allCtrlCVDict = {}
    allCtrls = rigUtil.listSetMembers(rigSettings.NODENAME_DICT["allCtrlSet"],["transform"],True)

    for ctrl in allCtrls:
        cvPositions,degree,form,knots = rigUtil.readCurveShape(ctrl,"object")
        allCtrlCVDict[ctrl] = {"cv":cvPositions}
    
    ##export for json
    filePath = setupDataPath + rigSettings.POSTFILENAME_DICT["ctrlShape"]
    rigUtil.writeJSON(filePath,allCtrlCVDict,indent = 4)

def importAllCtrlCV(setupDataPath):
    ##import from json
    filePath = setupDataPath + rigSettings.POSTFILENAME_DICT["ctrlShape"]
    allCtrlCVDict = rigUtil.readJSON(filePath)

    for ctrl in list(allCtrlCVDict.keys()):
        if cmds.objExists(ctrl) ==False:
            continue
        
        cvPositions = allCtrlCVDict[ctrl]["cv"]
        curCvPositions,degree,form,knots = rigUtil.readCurveShape(ctrl)
        
        if len(cvPositions) != len(curCvPositions):
            continue
        
        cvnum = len(cmds.ls(ctrl + ".cv[*]",fl =True))
        for i in range(0,cvnum):
            cmds.xform(ctrl + ".cv["+str(i)+"]",os = True,t = cvPositions[i])
    
def importAllCtrlValues(setupDataPath):
    filePath = setupDataPath + rigSettings.POSTFILENAME_DICT["ctrlValue"]
    allCtrlValueDict = rigUtil.readJSON(filePath)

    for ctrl in list(allCtrlValueDict.keys()):
        attrsDict = allCtrlValueDict[ctrl]

        for attr in attrsDict.keys():
            if cmds.objExists(ctrl + "." + attr) ==False:
                continue
            if cmds.getAttr(ctrl + "." + attr,lock =True):
                print(">>" + ctrl + "."+ attr + "  is locked")
                continue
            try:
                cmds.setAttr(ctrl + "." + attr, attrsDict[attr])
            except:
                print(">>" + ctrl + "."+ attr + "  cant setValue")

def importPostTransform(setupDataPath):
    filePath = setupDataPath + rigSettings.POSTFILENAME_DICT["transform"]
    postTransformDict = rigUtil.readJSON(filePath)

    for attr in list(postTransformDict.keys()):
        value = postTransformDict[attr]

        if attr == "scale":
            attr = rigUtil.addPrefix("post",attr)
                    
        rigUtil.setAttrState([rigSettings.NODENAME_DICT["ctrlWorldSpace"]],[attr],False,False,False)
        cmds.setAttr(rigSettings.NODENAME_DICT["ctrlWorldSpace"]+"." + attr,value)
        rigUtil.setAttrState([rigSettings.NODENAME_DICT["ctrlWorldSpace"]],[attr],True,True,True)

def restoreEditsProcess(setupDataPath,restoreCtrlValue,restoreCtrlSahpe,restorePostTransform):
    if restoreCtrlValue:
        importAllCtrlValues(setupDataPath)
    
    if restoreCtrlSahpe:
        importAllCtrlCV(setupDataPath)

    if restorePostTransform:
        importPostTransform(setupDataPath)
##-------------------------------------------------------------------------------
## for cleanUpProcess
##-------------------------------------------------------------------------------
def setCtrlDefaultValues(nodes):
    for node in nodes:
        unKeyableAttrs = cmds.listAttr(node,channelBox =True) or []
        keyableAttrs = cmds.listAttr(node,keyable =True,scalar =True) or []
        UDAttrs = cmds.listAttr(node,userDefined =True,scalar =True) or []

        attrs = []
        attrs.extend(unKeyableAttrs)
        attrs.extend(keyableAttrs)

        for attr in attrs:
            if attr in UDAttrs:
                if len(cmds.attributeQuery(attr, node = node, listChildren =True) or []) != 0:
                    continue

                cmds.addAttr(node + "."+attr,edit=True,dv = cmds.getAttr(node+"."+attr))

        matrix = cmds.getAttr(node + ".matrix")
        worldMatrix = cmds.getAttr(node + ".worldMatrix[0]")

        if cmds.attributeQuery("defaultMatrix", node = node, exists =True) == False:
            cmds.addAttr(node, dt = "matrix", ln = "defaultMatrix")
        
        if cmds.attributeQuery("defaultWorldMatrix", node = node, exists =True) == False:
            cmds.addAttr(node, dt = "matrix", ln = "defaultWorldMatrix")

        cmds.setAttr(node + ".defaultMatrix", matrix, type = "matrix")
        cmds.setAttr(node + ".defaultWorldMatrix", worldMatrix, type = "matrix")

def lockNodeAttrs(nodes,ignorAttrs):
    for node in nodes:
        unKeyableAttrs = cmds.listAttr(node,channelBox =True) or []
        keyableAttrs = cmds.listAttr(node,keyable =True) or []
        enableAttrs = cmds.listAttr(node,settable =True,visible =True,write =True,output =True,scalarAndArray =True) or []
        targetAttrs = list(set(enableAttrs) - (set(keyableAttrs) | set(ignorAttrs) | set(unKeyableAttrs)))        
        rigUtil.setAttrState([node],targetAttrs,lock=True,hide=True,keyable=False) 
        rigUtil.setAttrState([node],keyableAttrs,lock=True,hide=False,keyable=False)


def cleanUpProcess():
    ##非表示を設定
    cmds.setAttr(rigSettings.NODENAME_DICT["setupGP"] + ".v", False)
    cmds.setAttr(rigSettings.NODENAME_DICT["jointGP"] + ".v", False)
    
    ##階層下のノードを分類して取得
    nodeDict = rigUtil.listNodes(rigSettings.NODENAME_DICT["rootGP"],rigSettings.NODENAME_DICT["allCtrlSet"])

    ##コントローラーノードに デフォルトmatrix / worldMatrixを追加します
    ##コントローラーノードのアトリビュートにdefaultValueを設定します
    setCtrlDefaultValues(nodeDict["ctrl"])

    ignorAttrs = ["translate","rotate","scale","selectionChildHighlighting"]
    rigUtil.setAttrState([rigSettings.NODENAME_DICT["scaleCtrl"]],["postScale"],lock=True,hide=True,keyable=False)
    
    ##IKJoint
    lockNodeAttrs(nodeDict["IKJoints"],ignorAttrs)
    ##IKJoint はrotateを unKyeable　に設定
    rigUtil.setAttrState(nodeDict["IKJoints"],["rotate","rotateX","rotateY","rotateZ"],lock=False,hide=False,keyable=False)
    
    ##コンストレインノード
    ##アウトライナー内で非表示
    for node in nodeDict["constraint"]:
        cmds.setAttr(node + ".hiddenInOutliner",True)

    lockNodeAttrs(nodeDict["constraint"],ignorAttrs)

    ##それ以外のノード
    lockNodeAttrs(nodeDict["other"],ignorAttrs)


##-------------------------------------------------------------------------------
## for mainProcess
##-------------------------------------------------------------------------------
def setupMainProcess(modelFilePath,skeletonFilePath,setupDataPath,globalScale,restoreCtrlValue,restoreCtrlSahpe,restorePostTransform,applyCleanup):

    ##preProcess 001~099
    preProcess(modelFilePath,skeletonFilePath,globalScale,setupDataPath)

    ##setupProcess 100~899
    setupProcess(globalScale,setupDataPath)

    ##900~ 
    ##998
    ##postProcess 
    restoreEditsProcess(setupDataPath,restoreCtrlValue,restoreCtrlSahpe,restorePostTransform)

    #999
    ##cleanUpProcess 
    if applyCleanup:
        cleanUpProcess()